//20100127 - Alain Brobecker
#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */
#define TRUE 1
#define FALSE 0

int NbPawns=8;
int Position[8];
int Possibilities[8],SubsetPossibilities[8];
char Board[8][8];

/* for debugging */
void PrintPossibilities() {
  int Pawn,c;
  char d;
  for(Pawn=0;Pawn<NbPawns;Pawn++) {
    printf("P%c%d: ",'a'+Position[Pawn]/6,(Position[Pawn]%6)+2);    
    d='a';
    for(c=1;c<256;c=c<<1) {
      if((Possibilities[Pawn] & c)==0) { printf("."); } else { printf("%c",d); }
      d++;
    }
    printf("  ");
    d='a';
    for(c=1;c<256;c=c<<1) {
      if((SubsetPossibilities[Pawn] & c)==0) { printf("."); } else { printf("%c",d); }
      d++;
    }
    printf("\n");
  }
  printf("\n");
}

int CountBits(int Byte) {
  int c,NbBits=0;
  for(c=1;c<=128;c=c<<1) {
    if((Byte & c)!=0) { NbBits++; }
  }
  return(NbBits);
}

void PrintBoard() {
  int x,y,Pawn,c;
  for(y=0;y<8;y++) {
    for(x=0;x<8;x++) {
      Board[x][y]='.';
    }
  }
  for(Pawn=0;Pawn<NbPawns;Pawn++) {
    x=Position[Pawn]/6;
    y=Position[Pawn]%6+1;
    if(CountBits(Possibilities[Pawn])==1) {
      c=0;
      while((Possibilities[Pawn] & (1<<c))==0) { c++; }
      Board[x][y]='a'+c;
    } else {
      Board[x][y]='O';
    }
  }
  for(y=7;y>=0;y--) {
    for(x=0;x<8;x++) {
      printf("%c",Board[x][y]);
    }
    printf("\n");
  }
}

void PrintFEN() {
  int x,y,nb;
  for(y=7;y>=0;y--) {
    nb=0;
    for(x=0;x<8;x++) {
      if(Board[x][y]=='.') {
        nb++;
      } else {
        if(nb!=0) { printf("%d",nb); }
        printf("P"); //!!!NORMALLY printf("%c",Board[x][y]);
        nb=0;
      }
    }
//!!!NORMALLY    if(nb!=0) { printf("%d",nb); }
    if(nb!=0) { if(y==0){ printf("K6k"); }else{ printf("%d",nb); } }
    if(y!=0) { printf("/"); }
  }
  printf("\n");
}

/******************************************************************************
*************************                       *******************************
*************************   HOW MANY CAPTURES   *******************************
*************************                       *******************************
******************************************************************************/
/* HowManyCaptures() needs NbPawns==8, Position[8] and Possibilities[8] as
   global variables (see below for positions' numbering from white's viewpoint).
   Possibilities[0]==5 means that pawn at Position[0] comes from a or c files
   (5==%00000101==a.c.....).
   The 6*8 board is numbered the following way for white:
     7 | 5 11 17 23 29 35 41 47
     6 | 4 10 16 22 28 34 40 46
     5 | 3  9 15 21 27 33 39 45
     4 | 2  8 14 20 26 32 38 44
     3 | 1  7 13 19 25 31 37 43
     2 | 0  6 12 18 24 30 36 42
     --+-----------------------
       | a  b  c  d  e  f  g  h
   Below is the table of possible origins for a pawn at a given square: */
int PossibleOrigins[48] = {
   1,  3,  7, 15, 31, 63,  // .......a  ......ba  .....cba  ....dcba  ...edcba  ..fedcba
   2,  7, 15, 31, 63,127,  // ......b.  .....cba  ....dcba  ...edcba  ..fedcba  .gfedcba
   4, 14, 31, 63,127,255,  // .....c..  ....dcb.  ...edcba  ..fedcba  .gfedcba  hgfedcba
   8, 28, 62,127,255,255,  // ....d...  ...edc..  ..fedcb.  .gfedcba  hgfedcba  hgfedcba
  16, 56,124,254,255,255,  // ...e....  ..fed...  .gfedc..  hgfedcb.  hgfedcba  hgfedcba 
  32,112,248,252,254,255,  // ..f.....  .gfh....  hgfed...  hgfedc..  hgfedcb.  hgfedcba
  64,224,240,248,252,254,  // .g......  hgf.....  hgfe....  hgfed...  hgfedc..  hgfedcb.
 128,192,224,240,248,252}; // h.......  hg......  hgf.....  hgfe....  hgfed...  hgfedc..

int Mask,MixedPossibilities,NotMixedPossibilities;
int t1,t2,t3,Pawn,NbCaptures;
int NoChange; /* booleans */

int HowManyCaptures() {
  // (bubble) sort the pawns by position
  for(t1=0;t1<7;t1++) {
    for(t2=7;t2>t1;t2--) {
      if(Position[t2-1]>Position[t2]) { t3=Position[t2-1]; Position[t2-1]=Position[t2]; Position[t2]=t3; }
    }
  }

  for(Pawn=0;Pawn<NbPawns;Pawn++) { Possibilities[Pawn]=PossibleOrigins[Position[Pawn]]; }
  do{
    NoChange=TRUE;
    // Check for every N-uple of Pawns if a OR of all their possibilities
    // contains exactly N bits, in which case we must kill those possibilities
    // for the other Pawns. If it contains less than N bits then the position
    // is illegal. At the end the subsets will have empty intersection.
    for(Mask=(1<<NbPawns)-1;Mask>0;Mask--) {
      MixedPossibilities=0;
      for(Pawn=0;Pawn<NbPawns;Pawn++) {
        // Is the pawn in the N-uple?
        if(((1<<Pawn) & Mask)!=0) {
          MixedPossibilities=MixedPossibilities | Possibilities[Pawn];
        }
      }
      t1=CountBits(MixedPossibilities);
      t2=CountBits(Mask);
      if(t1<t2) {
        return(-1); // Position is illegal if N pawns have less than N possibilities
      } else if(t1==t2) {
        NotMixedPossibilities=~MixedPossibilities;
        // remove those possibilities for the pawns that were not in the N-uple
        for(Pawn=0;Pawn<NbPawns;Pawn++) {
          if(((1<<Pawn) & Mask)==0) {
            t1=Possibilities[Pawn];
            t2=t1 & NotMixedPossibilities;
            if(t1!=t2) {
              Possibilities[Pawn]=t2;
              NoChange=FALSE;
            }
          }
        }
      }
    }
  } while(NoChange==FALSE);

  // Create the subsets
  Mask=0; // to say if a Pawn has been seen or not
  SubsetPossibilities[NbPawns-1]=Possibilities[NbPawns-1];
  for(Pawn=0;Pawn<NbPawns-1;Pawn++) {
    if((Mask & (1<<Pawn))==0) {
      MixedPossibilities=Possibilities[Pawn]; // will contain all possible origins for this subset
      for(t1=Pawn+1;t1<NbPawns;t1++) {
        if((Possibilities[t1] & MixedPossibilities)!=0) {
          MixedPossibilities=MixedPossibilities | Possibilities[t1];
        }
      }
      for(t1=Pawn;t1<NbPawns;t1++) {
        if((Possibilities[t1] & MixedPossibilities)!=0) {
          Mask=Mask | (1<<t1);
          SubsetPossibilities[t1]=MixedPossibilities;
        }
      }
    }
  }
  
  // Count the nb of captures
  NbCaptures=0;
  Mask=255; // valid origins
  for(Pawn=0;Pawn<NbPawns;Pawn++) {
    t1=Position[Pawn]/6; // column of the pawn
    t2=0;
    while(((1<<t2) & Mask & SubsetPossibilities[Pawn])==0) { t2++; }
    Mask=Mask & (~(1<<t2)); //remove this possibility for other pawns
    t3=t2-t1;
    if(t3<0) { t3=-t3; }
    NbCaptures+=t3;
  }
 
  if(NbCaptures>15) { return(-NbCaptures); }
  return(NbCaptures);
}

/**** MAIN *******************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  int c,x,y;
  if(argc!=9) { printf("usage: 8Pawns a2 b2 b4 b5 b6 c7 d6 e4\n"); exit(-1); }
  for(c=1;c<9;c++) {
    x=argv[c][0]-'a'; if((x<0) || (x>7)) { printf("%s is incorrect\n",argv[c]); exit(-1); }
    y=argv[c][1]-'2'; if((y<0) || (y>5)) { printf("%s is incorrect\n",argv[c]); exit(-1); }
    Position[c-1]=x*6+y;
  }
  printf("Number of captures>=%d\n",HowManyCaptures());
  PrintBoard(); PrintFEN(); printf("\n");
  PrintPossibilities();
  exit(0);
}
